@@ -5,8 +5,8 @@
 ;; Author: Sergey Kostyaev <sskostyaev@gmail.com>
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
-;; Package-Requires: ((emacs "28.1") (llm "0.22.0") (spinner "1.7.4") (transient "0.7") (compat "29.1"))
-;; Version: 0.13.11
+;; Package-Requires: ((emacs "28.1") (llm "0.22.0") (spinner "1.7.4") (transient "0.7") (compat "29.1") (posframe "1.4.0"))
+;; Version: 1.0.3
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -41,6 +41,7 @@
 (require 'spinner)
 (require 'transient)
 (require 'compat)
+(require 'posframe)
 (eval-when-compile (require 'rx))
 
 (defgroup ellama nil
@@ -133,6 +134,7 @@
     (define-key map (kbd "i w") 'ellama-improve-wording)
     (define-key map (kbd "i g") 'ellama-improve-grammar)
     (define-key map (kbd "i c") 'ellama-improve-conciseness)
+    (define-key map (kbd "P") 'ellama-proofread)
     ;; make
     (define-key map (kbd "m l") 'ellama-make-list)
     (define-key map (kbd "m t") 'ellama-make-table)
@@ -143,6 +145,7 @@
     (define-key map (kbd "a l") 'ellama-ask-line)
     (define-key map (kbd "a s") 'ellama-ask-selection)
     ;; text
+    (define-key map (kbd "w") 'ellama-write)
     (define-key map (kbd "t t") 'ellama-translate)
     (define-key map (kbd "t b") 'ellama-translate-buffer)
     (define-key map (kbd "t c") 'ellama-complete)
@@ -155,6 +158,7 @@
     (define-key map (kbd "x f") 'ellama-context-add-file)
     (define-key map (kbd "x s") 'ellama-context-add-selection)
     (define-key map (kbd "x i") 'ellama-context-add-info-node)
+    (define-key map (kbd "x r") 'ellama-context-reset)
     ;; provider
     (define-key map (kbd "p s") 'ellama-provider-select)
     map)
@@ -256,6 +260,16 @@ You are a summarizer. You write a summary of the input **IN THE SAME LANGUAGE AS
   :group 'ellama
   :type 'string)
 
+(defcustom ellama-write-prompt-template "<SYSTEM>
+Write text, based on provided context and instruction. Do not add any explanation or acknowledgement, just follow instruction.
+</SYSTEM>
+<INSTRUCTION>
+%s
+</INSTRUCTION>"
+  "Prompt template for `ellama-write'."
+  :group 'ellama
+  :type 'string)
+
 (defcustom ellama-improve-grammar-prompt-template "improve grammar and spelling"
   "Prompt template for `ellama-improve-grammar'."
   :group 'ellama
@@ -266,6 +280,11 @@ You are a summarizer. You write a summary of the input **IN THE SAME LANGUAGE AS
   :group 'ellama
   :type 'string)
 
+(defcustom ellama-proofread-prompt-template "proofread"
+  "Prompt template for `ellama-proofread'."
+  :group 'ellama
+  :type 'string)
+
 (defcustom ellama-improve-conciseness-prompt-template "make it as simple and concise as possible"
   "Prompt template for `ellama-improve-conciseness'."
   :group 'ellama
@@ -286,7 +305,7 @@ You are a summarizer. You write a summary of the input **IN THE SAME LANGUAGE AS
   :group 'ellama
   :type 'string)
 
-(defcustom ellama-code-add-prompt-template "Context: \n```\n%s\n```\nBased on this context, %s, only output the result in format ```\n...\n```\nWrite all the code in single code block."
+(defcustom ellama-code-add-prompt-template "Based on context, %s, only output the result in format ```\n...\n```\nWrite all the code in single code block."
   "Prompt template for `ellama-code-add'."
   :group 'ellama
   :type 'string)
@@ -738,8 +757,6 @@ EXTRA contains additional information."
   "Generate name for ellama ACTION by PROVIDER according to PROMPT."
   (ellama--fix-file-name (funcall ellama-naming-scheme provider action prompt)))
 
-(defvar ellama--new-session-context nil)
-
 (defun ellama-get-nick-prefix-for-mode ()
   "Return preferred header prefix char based om the current mode.
 Defaults to #, but supports `org-mode'.  Depends on `ellama-major-mode'."
@@ -754,6 +771,9 @@ Defaults to md, but supports org.  Depends on \"ellama-major-mode.\""
   (cond ((provided-mode-derived-p ellama-major-mode 'org-mode) "org")
         (t "md")))
 
+(defvar ellama--global-context nil
+  "Global context.")
+
 (defun ellama-new-session (provider prompt &optional ephemeral)
   "Create new ellama session with unique id.
 Provided PROVIDER and PROMPT will be used in new session.
@@ -779,15 +799,14 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	      ellama--current-session)))
 	 (session (make-ellama-session
 		   :id id :provider provider :file file-name
-		   :context (if previous-session
-				(ellama-session-context previous-session)
-			      ellama--new-session-context)))
+		   :context (or (when previous-session
+				  (ellama-session-context previous-session))
+				ellama--global-context)))
 	 (buffer (if file-name
 		     (progn
 		       (make-directory ellama-sessions-directory t)
 		       (find-file-noselect file-name))
 		   (get-buffer-create id))))
-    (setq ellama--new-session-context nil)
     (setq ellama--current-session-id id)
     (puthash id buffer ellama--active-sessions)
     (with-current-buffer buffer
@@ -915,9 +934,8 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	       :provider (ellama-session-provider session)
 	       :file (ellama-session-file session)
 	       :prompt (ellama-session-prompt session)
-	       :context ellama--new-session-context
+	       :context ellama--global-context
 	       :extra extra)))
-      (setq ellama--new-session-context nil)
       (setq ellama--current-session-id (ellama-session-id ellama--current-session))
       (puthash (ellama-session-id ellama--current-session)
 	       buffer ellama--active-sessions)
@@ -1000,6 +1018,17 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (remhash id ellama--active-sessions)
     (puthash new-id buffer ellama--active-sessions)))
 
+(defvar ellama--context-buffer " *ellama-context*")
+
+;;;###autoload
+(defun ellama-context-reset ()
+  "Clear global context."
+  (interactive)
+  (setq ellama--global-context nil)
+  (with-current-buffer ellama--context-buffer
+    (erase-buffer))
+  (posframe-hide ellama--context-buffer))
+
 ;; Context elements
 
 (defclass ellama-context-element () ()
@@ -1011,16 +1040,51 @@ If EPHEMERAL non nil new session will not be associated with any file."
 (cl-defgeneric ellama-context-element-extract (element)
   "Extract the content of the context ELEMENT.")
 
+(cl-defgeneric ellama-context-element-display (element)
+  "Display the context ELEMENT.")
+
 (cl-defgeneric ellama-context-element-format (element mode)
   "Format the context ELEMENT for the major MODE.")
 
+(defcustom ellama-context-poshandler 'posframe-poshandler-frame-top-center
+  "Position handler for displaying context buffer."
+  :group 'ellama
+  :type 'function)
+
+(defcustom ellama-context-border-width 1
+  "Border width for the context buffer."
+  :group 'ellama
+  :type 'integer)
+
+(defcustom ellama-context-element-padding-size 20
+  "Padding size for context elements."
+  :group 'ellama
+  :type 'integer)
+
 (cl-defmethod ellama-context-element-add ((element ellama-context-element))
   "Add the ELEMENT to the Ellama context."
   (if-let* ((id ellama--current-session-id)
 	    (session (with-current-buffer (ellama-get-session-buffer id)
 		       ellama--current-session)))
-      (push element (ellama-session-context session))
-    (push element ellama--new-session-context)))
+      (push element (ellama-session-context session)))
+  (push element ellama--global-context)
+  (get-buffer-create ellama--context-buffer t)
+  (with-current-buffer ellama--context-buffer
+    (erase-buffer)
+    (insert (format
+	     "context: %s"
+	     (string-join
+	      (mapcar
+	       (lambda (el)
+		 (string-pad
+		  (ellama-context-element-display el) ellama-context-element-padding-size))
+	       ellama--global-context)
+	      "  "))))
+  (posframe-show
+   ellama--context-buffer
+   :poshandler ellama-context-poshandler
+   :internal-border-width ellama-context-border-width))
+
 
 ;; Buffer context element
 
@@ -1035,6 +1099,12 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (with-current-buffer name
       (buffer-substring-no-properties (point-min) (point-max)))))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-buffer))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    name))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-buffer) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1063,6 +1133,12 @@ If EPHEMERAL non nil new session will not be associated with any file."
       (insert-file-contents name)
       (buffer-substring-no-properties (point-min) (point-max)))))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-file))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    (file-name-nondirectory name)))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-file) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1091,6 +1167,12 @@ If EPHEMERAL non nil new session will not be associated with any file."
       (info name (current-buffer))
       (buffer-substring-no-properties (point-min) (point-max)))))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-info-node))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    (format "(info \"%s\")" name)))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-info-node) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1122,6 +1204,21 @@ If EPHEMERAL non nil new session will not be associated with any file."
   "Extract the content of the context ELEMENT."
   (oref element content))
 
+(defcustom ellama-text-display-limit 15
+  "Limit for text display in context elements."
+  :group 'ellama
+  :type 'integer)
+
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-text))
+  "Display the context ELEMENT."
+  (with-slots (content) element
+    (format "\"%s\"" (concat
+		      (string-limit
+		       content
+		       ellama-text-display-limit)
+		      "..."))))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-text) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1147,6 +1244,18 @@ If EPHEMERAL non nil new session will not be associated with any file."
   "Extract the content of the context ELEMENT."
   (oref element content))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-webpage-quote))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    name))
+
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-webpage-quote))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    name))
+
 (defun ellama--quote-buffer (quote)
   "Return buffer name for QUOTE."
   (let* ((buf-name (concat (make-temp-name "*ellama-quote-") "*"))
@@ -1209,6 +1318,12 @@ If EPHEMERAL non nil new session will not be associated with any file."
   "Extract the content of the context ELEMENT."
   (oref element content))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-info-node-quote))
+  "Display the context ELEMENT."
+  (with-slots (name) element
+    (format "(info \"%s\")" name)))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-info-node-quote) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1255,6 +1370,12 @@ If EPHEMERAL non nil new session will not be associated with any file."
   "Extract the content of the context ELEMENT."
   (oref element content))
 
+(cl-defmethod ellama-context-element-display
+  ((element ellama-context-element-file-quote))
+  "Display the context ELEMENT."
+  (with-slots (path) element
+    (file-name-nondirectory path)))
+
 (cl-defmethod ellama-context-element-format
   ((element ellama-context-element-file-quote) (mode (eql 'markdown-mode)))
   "Format the context ELEMENT for the major MODE."
@@ -1415,15 +1536,18 @@ If EPHEMERAL non nil new session will not be associated with any file."
 
 (defun ellama--prompt-with-context (prompt)
   "Add context to PROMPT for sending to llm."
-  (if-let* ((session ellama--current-session)
-	    (context (ellama-session-context session)))
-      (concat (string-join
-	       (cons "Context:"
-		     (mapcar #'ellama-context-element-extract context))
-	       "\n")
-	      "\n\n"
-	      prompt)
-    prompt))
+  (let* ((session ellama--current-session)
+	 (context (or (when session
+			(ellama-session-context session))
+		      ellama--global-context)))
+    (if context
+	(concat (string-join
+		 (cons "Context:"
+		       (mapcar #'ellama-context-element-extract context))
+		 "\n")
+		"\n\n"
+		prompt)
+      prompt)))
 
 (defun ellama-chat-buffer-p (buffer)
   "Return non-nil if BUFFER is an ellama chat buffer."
@@ -2080,6 +2204,17 @@ ARGS contains keys for fine control.
     (ellama-context-add-buffer (current-buffer)))
   (ellama-chat ellama-code-review-prompt-template nil :provider ellama-coding-provider))
 
+;;;###autoload
+(defun ellama-write (instruction)
+  "Write text based on context and INSTRUCTION at point."
+  (interactive "sInstruction: ")
+  (when (region-active-p)
+    (ellama-context-add-selection))
+  (ellama-stream (format ellama-write-prompt-template instruction)
+		 :point (point)
+		 :filter (when (derived-mode-p 'org-mode)
+			   #'ellama--translate-markdown-to-org-filter)))
+
 ;;;###autoload
 (defun ellama-change (change &optional edit-template)
   "Change selected text or text in current buffer according to provided CHANGE.
@@ -2118,6 +2253,14 @@ prefix (\\[universal-argument]), prompt the user to amend the template."
   (interactive "p")
   (ellama-change ellama-improve-wording-prompt-template edit-template))
 
+;;;###autoload
+(defun ellama-proofread (&optional edit-template)
+  "Proofread the currently selected region or buffer.
+When the value of EDIT-TEMPLATE is 4, or with one `universal-argument' as
+prefix (\\[universal-argument]), prompt the user to amend the template."
+  (interactive "p")
+  (ellama-change ellama-proofread-prompt-template edit-template))
+
 ;;;###autoload
 (defun ellama-improve-conciseness (&optional edit-template)
   "Make the text of the currently selected region or buffer concise and simple.
@@ -2187,23 +2330,19 @@ prefix (\\[universal-argument]), prompt the user to amend the template."
 
 ;;;###autoload
 (defun ellama-code-add (description)
-  "Add new code according to DESCRIPTION.
-Code will be generated with provided context from selected region or current
-buffer."
+  "Generate and insert new code based on DESCRIPTION.
+This function prompts the user to describe the code they want to generate.
+If a region is active, it includes the selected text as context for code
+generation."
   (interactive "sDescribe the code to be generated: ")
-  (let* ((beg (if (region-active-p)
-		  (region-beginning)
-		(point-min)))
-	 (end (if (region-active-p)
-		  (region-end)
-		(point-max)))
-	 (text (buffer-substring-no-properties beg end)))
-    (ellama-stream
-     (format
-      ellama-code-add-prompt-template
-      text description)
-     :provider ellama-coding-provider
-     :filter #'ellama--code-filter)))
+  (when (region-active-p)
+    (ellama-context-add-selection))
+  (ellama-stream
+   (format
+    ellama-code-add-prompt-template
+    description)
+   :provider ellama-coding-provider
+   :filter #'ellama--code-filter))
 
 
 ;;;###autoload
@@ -2343,20 +2482,25 @@ Call CALLBACK on result list of strings.  ARGS contains keys for fine control.
      (lambda (err)
        (user-error err)))))
 
+(defun ellama-get-ollama-model-name ()
+  "Get ollama model name from installed locally."
+  (interactive)
+  (completing-read
+   "Select ollama model: "
+   (mapcar (lambda (s)
+	     (car (split-string s)))
+	   (seq-drop
+	    (process-lines
+	     (executable-find ellama-ollama-binary) "ls")
+	    1))))
+
 (defun ellama-get-ollama-local-model ()
   "Return llm provider for interactively selected ollama model."
   (interactive)
   (declare-function llm-ollama-p "ext:llm-ollama")
   (declare-function llm-ollama-host "ext:llm-ollama")
   (declare-function llm-ollama-port "ext:llm-ollama")
-  (let ((model-name
-	 (completing-read "Select ollama model: "
-			  (mapcar (lambda (s)
-				    (car (split-string s)))
-				  (seq-drop
-				   (process-lines
-				    (executable-find ellama-ollama-binary) "ls")
-				   1))))
+  (let ((model-name (ellama-get-ollama-model-name))
 	(host (when (llm-ollama-p ellama-provider)
 		(llm-ollama-host ellama-provider)))
 	(port (when (llm-ollama-p ellama-provider)
@@ -2367,6 +2511,121 @@ Call CALLBACK on result list of strings.  ARGS contains keys for fine control.
       (make-llm-ollama
        :chat-model model-name :embedding-model model-name))))
 
+(defvar ellama-transient-ollama-model-name "")
+(defvar ellama-transient-temperature 0.7)
+(defvar ellama-transient-context-length 4096)
+(defvar ellama-transient-host nil)
+(defvar ellama-transient-port nil)
+
+(transient-define-suffix ellama-transient-set-ollama-model ()
+  "Set ollama model name."
+  (interactive)
+  (setq ellama-transient-ollama-model-name (ellama-get-ollama-model-name)))
+
+(transient-define-suffix ellama-transient-set-temperature ()
+  "Set temperature value."
+  (interactive)
+  (setq ellama-transient-temperature (read-number "Enter temperature: ")))
+
+(transient-define-suffix ellama-transient-set-context-length ()
+  "Set context length."
+  (interactive)
+  (setq ellama-transient-context-length (read-number "Enter context length: ")))
+
+(transient-define-suffix ellama-transient-set-host ()
+  "Set host address."
+  (interactive)
+  (setq ellama-transient-host (read-string "Enter host: ")))
+
+(transient-define-suffix ellama-transient-set-port ()
+  "Set port number."
+  (interactive)
+  (setq ellama-transient-port (read-number "Enter port: ")))
+
+(defvar ellama-provider-list '(ellama-provider
+			       ellama-coding-provider
+			       ellama-translation-provider
+			       ellama-extraction-provider
+			       ellama-summarization-provider
+			       ellama-naming-provider)
+  "List of ollama providers.")
+
+(transient-define-suffix ellama-transient-model-get-from-provider ()
+  "Fill transient model from provider."
+  (interactive)
+  (ellama-fill-transient-ollama-model
+   (eval (read
+	  (completing-read "Select provider: "
+			   (mapcar #'prin1-to-string ellama-provider-list))))))
+
+(transient-define-suffix ellama-transient-set-provider ()
+  "Set transient model to provider."
+  (interactive)
+  (set (read
+	(completing-read "Select provider: "
+			 (mapcar #'prin1-to-string ellama-provider-list)))
+       (ellama-construct-ollama-provider-from-transient)))
+
+(transient-define-prefix ellama-select-ollama-model ()
+  "Select ollama model."
+  [["Model"
+    ("f" "Load from provider" ellama-transient-model-get-from-provider
+     :transient t)
+    ("m" "Set Model" ellama-transient-set-ollama-model
+     :transient t
+     :description (lambda () (format "Model (%s)" ellama-transient-ollama-model-name)))
+    ("t" "Set Temperature" ellama-transient-set-temperature
+     :transient t
+     :description (lambda () (format "Temperature (%.2f)" ellama-transient-temperature)))
+    ("c" "Set Context Length" ellama-transient-set-context-length
+     :transient t
+     :description (lambda () (format "Context Length (%d)" ellama-transient-context-length)))
+    ("S" "Set provider" ellama-transient-set-provider
+     :transient t)
+    ("s" "Set provider and quit" ellama-transient-set-provider)]
+   ["Connection"
+    ("h" "Set Host" ellama-transient-set-host
+     :transient t
+     :description (lambda () (if ellama-transient-host
+				 (format "Host (%s)" ellama-transient-host)
+			       "Host")))
+    ("p" "Set Port" ellama-transient-set-port
+     :transient t
+     :description (lambda () (if ellama-transient-port
+				 (format "Port (%s)" ellama-transient-port)
+			       "Port")))]
+   ["Quit" ("q" "Quit" transient-quit-one)]])
+
+(defun ellama-fill-transient-ollama-model (provider)
+  "Set transient ollama model from PROVIDER."
+  (declare-function llm-ollama-p "ext:llm-ollama")
+  (declare-function llm-ollama-host "ext:llm-ollama")
+  (declare-function llm-ollama-port "ext:llm-ollama")
+  (declare-function llm-ollama-chat-model "ext:llm-ollama")
+  (declare-function llm-ollama-default-chat-temperature "ext:llm-ollama")
+  (declare-function llm-ollama-default-chat-non-standard-params "ext:llm-ollama")
+  (when (llm-ollama-p provider)
+    (setq ellama-transient-ollama-model-name (llm-ollama-chat-model provider))
+    (setq ellama-transient-temperature (or (llm-ollama-default-chat-temperature provider) 0.7))
+    (setq ellama-transient-host (llm-ollama-host provider))
+    (setq ellama-transient-port (llm-ollama-port provider))
+    (let* ((other-params (llm-ollama-default-chat-non-standard-params provider))
+	   (ctx-len (when other-params (alist-get
+					"num_ctx"
+					(seq--into-list other-params)
+					nil nil #'string=))))
+      (setq ellama-transient-context-length (or ctx-len 4096)))))
+
+(defun ellama-construct-ollama-provider-from-transient ()
+  "Make provider with ollama mode in transient menu."
+  (make-llm-ollama
+   :chat-model ellama-transient-ollama-model-name
+   :default-chat-temperature ellama-transient-temperature
+   :host ellama-transient-host
+   :port ellama-transient-port
+   :default-chat-non-standard-params
+   `[("num_ctx" . ,ellama-transient-context-length)]))
+
 (transient-define-prefix ellama-transient-code-menu ()
   "Code Commands."
   [["Code Commands"
@@ -2434,15 +2693,19 @@ Call CALLBACK on result list of strings.  ARGS contains keys for fine control.
     ("b" "Add Buffer" ellama-context-add-buffer)
     ("f" "Add File" ellama-context-add-file)
     ("s" "Add Selection" ellama-context-add-selection)
-    ("i" "Add Info Node" ellama-context-add-info-node)]
+    ("i" "Add Info Node" ellama-context-add-info-node)
+    ("r" "Context reset" ellama-context-reset)]
    ["Quit" ("q" "Quit" transient-quit-one)]])
 
 (transient-define-prefix ellama-transient-main-menu ()
   "Main Menu."
   [["Main"
     ("c" "Chat" ellama-chat)
+    ("w" "Write" ellama-write)
+    ("P" "Proofread" ellama-proofread)
     ("a" "Ask Commands" ellama-transient-ask-menu)
-    ("C" "Code Commands" ellama-transient-code-menu)]]
+    ("C" "Code Commands" ellama-transient-code-menu)
+    ("o" "Ollama model" ellama-select-ollama-model)]]
   [["Text"
     ("s" "Summarize Commands" ellama-transient-summarize-menu)
     ("i" "Improve Commands" ellama-transient-improve-menu)
